/*
 * PDM - Pixil display manager
 *
 * Copyright (C) 2003, Penguru Consulting, LLC
 * Written by:  Jordan Crouse (jordan@penguru.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

/* This is where all the config options are read and processed.  In the end,
   the dminfo_t * structure will be filled with everything you ever wanted
   to know about the PDM session.
*/

/* strcmpi()
   Case insensitive matching - like strcmp() but considers all alpha
   chars to be the same 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <libgen.h>

#include <sys/types.h>

#include "pdm.h"

#ifndef PDM_DEFAULT_DIR 
#define PDM_DEFAULT_DIR "/etc/pdm"
#endif

/* These functions help obstuficate the default directory.
   The default directory is either the value of PDM_DEFAULT_DIR or
   specified by the -d directory at the command line
*/

static char *pdm_dir = 0;

void SET_PDM_DIR(const char *dir) {
  if (pdm_dir) free(pdm_dir);
  pdm_dir = (dir) ? strdup(dir) : 0;
}

const char *GET_PDM_DIR(void) {
  if (pdm_dir) 
    return (const char *) pdm_dir;
  else
    return PDM_DEFAULT_DIR;
}

   
static int strcmpi(const char *astr, const char *bstr) {
  const char *a = astr, *b = bstr;
  
  for(; *a && *b; a++, b++) { 
    if (tolower(*a) == tolower(*b)) continue;
    return (tolower(*a) < tolower(*b)) ? -1 : 1;
  }

  if (*a == *b) return 0;
  return (*a) ? 1 : -1;
}
  
/* do_args()
   Process the argument string and put the results into a null terminated
   argv
*/

static int parse_cmdline(char **app, char ***ret, const char *argstr) {
  int argc = 0;
  char **argv = 0;
  int size;
  
  const char *s = argstr, *e = 0;

  /* The first item is the app. */
  for(e = s; *e && *e != ' '; e++);

  size = (int) (e - s);
  *app = calloc(size + 1, 1);
  strncpy(*app, s, size);
  
  /* The first argv[] is the basename of the app */
  argv = (char **) calloc(argc + 1, sizeof(char *));
  argv[argc++] = strdup(basename(*app));

  if (*e) {
    s = e + 1;
    
    while(*s) {
      for(e = s; *e && *e != ' '; e++);    
      argv = (char **) realloc(argv, (argc + 1) * sizeof(char *));
      size = (int) (e - s);
      argv[argc] = calloc(size + 1, 1);
      strncpy(argv[argc++], s, size);   
      s = e + 1;
    }
  }

  /* Null terminate the list */

  argv = (char **) realloc(argv, (argc + 1) * sizeof(char *));
  argv[argc] = 0;
  
  *ret = argv;

  return argc;
}

/* server_callback()
   Callback for the 'Server' keyword
*/

static int server_callback(int line, char *key, char *value, dminfo_t *info) {

  if (info->server) {
    fprintf(stderr, "%d: Server already specified.  Ignoring.\n", line);
    return 0;
  }

  info->argc = parse_cmdline(&info->server, &info->argv, value);
  return 0;
}

/* auth_callback()
   Callback for the 'Auth' keyword
*/
 
static int auth_callback(int line, char *key, char *value, dminfo_t *info) {

  if (info->flags & FLAG_AUTH) {
    fprintf(stderr, "%d: Auth already specified.  Ignoring.\n", line);
    return 0;
  }
    
  info->flags &= ~FLAG_AUTH;

  if (value && !strcmpi(value, "none")) 
    return 0;
  
  info->authplugin = load_auth_plugin(value);

  if (!info->authplugin) {
    fprintf(stderr, "Fatal:  Auth plugin '%s' is invalid on line %d.\n", value, line);
    return -1;
  }

  info->flags |= FLAG_AUTH;

  return 0;
}

/* user_callback()
   Callback for the 'User' keyword
*/

static int user_callback(int line, char *key, char *value, dminfo_t *info) {

  struct passwd *user = 0;

  if (info->flags & FLAG_AUTH) {
    fprintf(stderr, "%d: User is ignored when auth is turned on.  Ignoring.\n", line);
    return 0;
  }

  if (info->user) {
    fprintf(stderr, "%d: User already specified.  Ignoring.\n", line);
    return 0;
  }

  if (!*value) {
    fprintf(stderr, "%d: Error.  No User specified.\n", line);
    return -1;
  }

  /* Numbers are valid too */

  if (isdigit(*value))
    user = getpwuid(atoi(value));
  else
    user = getpwnam(value);

  if (!user) {
    fprintf(stderr, "%d: Error.  User '%s' doesn't exist\n", line, value);
    return -1;
  }  

  info->user = (struct passwd *) calloc(sizeof(struct passwd), 1);

  if (!info->user) {
    fprintf(stderr, "Memory error while allocating user data\n");
    return -1;
  }

  info->user->pw_name = user->pw_name ? strdup(user->pw_name) : 0;
  info->user->pw_passwd = user->pw_name ? strdup(user->pw_passwd) : 0;
  info->user->pw_uid = user->pw_uid;
  info->user->pw_gid = user->pw_gid;
  info->user->pw_gecos = user->pw_gecos ? strdup(user->pw_gecos) : 0;
  info->user->pw_dir = user->pw_dir ? strdup(user->pw_dir) : 0;
  info->user->pw_shell = user->pw_shell ? strdup(user->pw_shell) : 0;

  return 0;
}

/* session_callback()
   Callback for the 'Session' keyword
*/

static int session_callback(int line, char *key, char *value, dminfo_t *info) {

  if (info->session) {
    fprintf(stderr, "%d: Session already specified.  Ignoring.\n", line);
    return 0;
  }

  info->session = (value) ? strdup(value) : 0;
  return 0;
}

/* calibrate_callback()
   Callback for the 'TSCalibrate' keyword
*/

static int calibrate_callback(int line, char *key, char *value, dminfo_t *info) {

  if (info->flags & FLAG_CAL) {
    fprintf(stderr, "%d: Calibration already specified.  Ignoring.\n", line);
    return 0;
  }
  
  /* do a case insensitve comparision */
  info->flags &= ~FLAG_CAL;

  if (value && !strcmpi(value, "yes")) info->flags |= FLAG_CAL;
  return 0;
}

/* calbin_callback()
   Callback for the 'TSCalBinary' keyword
*/

static int calbin_callback(int line, char *key, char *value, dminfo_t *info) {

  if (!(info->flags & FLAG_CAL)) {
    fprintf(stderr, "%d: Calibration is not active.  Ignoring.\n", line);
    return 0;
  }

  info->cal.argc = parse_cmdline(&info->cal.bin, &info->cal.argv, value);
  return 0;
}

/* The list of keywords.  We don't do the typical 
   "NULL" terminated list here to save a bit of room.  
*/

#define KEYWORD_COUNT 6

struct {
  char *keyword;
  int (*callback)(int, char *, char *, dminfo_t *);
} keywords[KEYWORD_COUNT] = {
  { "Server", server_callback },
  { "Auth", auth_callback },
  { "User", user_callback },
  { "Session", session_callback },
  { "TSCalibrate", calibrate_callback },
  { "TSCalBinary", calbin_callback },
};

/*  do_config()
    Given a filename, process it and fill out the dminfo_t structure 
*/

int do_config(const char *filename, dminfo_t *info) {  

  int ret = -1, line = 0;

  FILE *stream = fopen(filename, "r");

  if (!stream) {
    fprintf(stderr, "Error:  Unable to open file %s\n", filename);
    return -1;
  }

  while(!feof(stream) && !ferror(stream)) {
    char buffer[128];
    char *key, *value;
    int i;

    if (!(key = fgets(buffer, sizeof(buffer) - 1, stream)))
      break;

    line++;

    /* strip the newline */
    if (buffer[strlen(buffer) - 1] == '\n') buffer[strlen(buffer) - 1] = 0;
    
    /* strip whitespace from the beginning */

    for ( ; *key && (*key == ' ' || *key == '\t'); key++);
    if (!*key || *key == '#') continue; /* Comment or blank space */
    
    for (value = key; *value && *value != '='; value++);

    if (!*value) {
      fprintf(stderr, "%d: Warning - Unknown or empty keyword %s.\n", line, key);
      continue;
    }
    *value++ = 0;

    for(i = 0; i < KEYWORD_COUNT; i++) {
      if (strcmp(keywords[i].keyword, key)) continue;

      if (keywords[i].callback) 
	if (keywords[i].callback(line, key, value, info) == -1) 
	  goto exit_loop;
      
      break;
    }
  }

  ret = 0;

 exit_loop:
  fclose(stream);
  return ret;
}

/* read_config()
   If the user provided a config file, use that, otherwise look for
   the config file in all the usual places (user homepage, then
   default directory)
*/

int read_config(char *userconfig, dminfo_t *info) {

  char buffer[128];
  int ret = -1;

  struct passwd *user = 0;

  /* If the user passed in a config file, read that no matter what */
  if (userconfig) return do_config(userconfig, info);

  /* Check to see if the user has a personal config */
  user = getpwuid(getuid());
  
  if (user) {
    snprintf(buffer, sizeof(buffer) - 1, "%s/.pdm/user.conf", user->pw_dir);
    ret = access(buffer, R_OK);
  }

  if (ret) {
    snprintf(buffer, sizeof(buffer) - 1, "%s/pdm.conf", GET_PDM_DIR());   
    ret = access(buffer, R_OK);
  }

  if (!ret) 
    return do_config(buffer, info);
    
  fprintf(stderr, "Error:  Could not find a config file.  Exiting.\n");
  return -1;
}


