/*
 * PDM - Pixil display manager
 *
 * Copyright (C) 2003, Penguru Consulting, LLC
 * Written by:  Jordan Crouse (jordan@penguru.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

/* Pixil Display Manager (PDM) 
   The PDM is a special utility that helps to simplify the process of
   starting Pixil (or any graphical environment for that matter).
   
   PDM is unique in that it is friendly to multiple graphics clients,
   it understand touchscreen calibration out of the box, and that it
   handles authinication through a plugin system.  
*/

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#include <time.h>

#include "pdm.h"

#define PDM_DEFAULT_SESSION "Default"

static int verbose = 0;

static inline void VERBOSE(char *fmt, ...) {
  va_list args;

  if (!verbose) return;

  va_start (args, fmt);
  vprintf(fmt, args);
  va_end(args);
}

/* This structure contains all the information read from the configuration
   file */

static dminfo_t dminfo;

/* spawn_app()
   Spawn the given application.  A simple fork() / execvp() is all that
   is needed here */

static pid_t spawn_app(char *app, char **argv) {

  pid_t pid = 0;

  if (!(pid = fork())) {
    VERBOSE("Trying to spawn app %s in pid '%d'\n", app, pid);
    execvp(app, argv);

    exit(-1);  
  }
  
  return pid;
}

/* spawn_session()
   Spawns a new session based on the given name under the authority
   of the given UID.
*/

static pid_t spawn_session(char *app, uid_t uid) {

  char *argv[2] = { 0, 0 };  
  pid_t pid = 0;
  
  /* A few last sanity checks */
  if (access(app, X_OK)) {
    fprintf(stderr, "Fatal:  '%s' is not an executable\n", app);
    return -1;
  }

  if (!(pid = fork())) {
    
    VERBOSE("Trying to spawn session %s in pid '%d'\n", app, pid);

    setsid();

    /* If a different UID was specified, then try to change to that UID */

    if (uid != getuid()) {
      if (seteuid(uid) == -1) {
	fprintf(stderr, "Error:  Unable to set the UID of the process.\n");
	exit(-1);
      }
    }

    argv[0] = app;
    execvp(app, argv);
    
    exit(-1);  /* Only if there is an error */
  }

  return pid;
}

static char *find_file(char *fmt, ...) {
  va_list ap;
  int size = 128;
  char *str = 0;
  
  if (!(str = calloc(size, 1))) return 0;
  
  while(1) {
    int n;

    va_start(ap, fmt);
    n = vsnprintf(str, size, fmt, ap);
    va_end(ap);

    if (n > -1 && n < size) break;
    if (n > -1) size = n + 1;
    
    if (!(str = realloc(str, size))) return 0;
  }

  if (!access(str, R_OK)) return str;
  free(str);
  return 0;
}

/* find_session()
   Given the user and session, find the file (either in the users home
   directory, or the default PDM location
*/

static char *find_session(struct passwd *user, const char *session) {

  char *file = 0;

  if (!session) return 0;

  if (user && user->pw_dir) {
    file = find_file("%s/.gendm/%s", user->pw_dir, session);
    if (file) return file;
  }

  file = find_file("%s/sessions/%s", GET_PDM_DIR(), session);
  if (file) return file;

  file = find_file("%s/sessions/%s", GET_PDM_DIR(), PDM_DEFAULT_SESSION);
  return file;
}

/* wait_for_child()
   A simple function, all we do is sit around and wait for a child to
   return.
*/

static pid_t wait_for_child(void) {
  int status;
  pid_t pid = wait(&status);
    
  fprintf(stderr, "Message:  Pid '%d' returned with status '%d'\n", 
	  pid, WEXITSTATUS(status));

  return pid;
}

static void kill_child(pid_t pid) {
  int status;

  kill(pid, SIGTERM);
  waitpid(pid, &status, 0);
}

int main(int argc, char **argv) {

  int ret = -1;
  char *configfile = 0;

  pid_t server_pid = 0;
  pid_t session_pid = 0;
  
  char *session = 0;
  struct passwd *curuser = 0;

  time_t timestamp = 0;
  int count = 0;

  printf("PDM Version 1.0 starting.\n");

  while(1) {
    signed char ch;
    int index = 0;

    static struct option options[] = {
      { "config", 1, 0, 'c' },
      { "dir", 1, 0, 'd' },
      { "verbose", 1, 0, 'v' },
      { 0, 0, 0, 0}
    };

    ch = getopt_long(argc, argv, "c:d:v", options, &index);
    if (ch == -1) break;

    switch(ch) {
    case 'c': configfile = optarg; break;
    case 'd': SET_PDM_DIR(optarg); break;
    case 'v': verbose = 1; break;
    }
  }

  memset(&dminfo, 0, sizeof(dminfo_t));

  /* If the config fails, we're over even before we begin */
  if (read_config(configfile, &dminfo) == -1) exit(-1);

  if (dminfo.session) {
    VERBOSE("Finding the session '%s'\n", dminfo.session);
    session = find_session(dminfo.user, dminfo.session);

    /* Its only fatal if auth isn't turned on.  Otherwise, its up to the plugin
       to decide what session to use 
    */

    if (!(dminfo.flags & FLAG_AUTH) && !session) {
      fprintf(stderr, "Fatal:  Session '%s' not found.\n", dminfo.session);
      goto gendm_done;
    }
  }

  VERBOSE("Spawning server '%s'\n", dminfo.server);

  /* Step 1 - we need to fire up the server */
  server_pid = spawn_app(dminfo.server, dminfo.argv);

  if (server_pid == -1) {
    fprintf(stderr, "Fatal: Unable to start the server.\n");  
    goto gendm_done;
  }

  /* Step 2 - If calibration is indicated, then start the calibration too */
   
  if (dminfo.flags & FLAG_CAL) {
    pid_t cal_pid;

    VERBOSE("Spawning calibration utility '%s'\n", dminfo.cal.bin);
    cal_pid = spawn_app(dminfo.cal.bin, dminfo.cal.argv);
    
    /* If the spawn was successful, then wait for the pid to return */

    if (cal_pid > 0) {
      pid_t reaped = wait_for_child();

      /* If the server died on us, then bail */

      if (reaped == server_pid) {
	fprintf(stderr, "Fatal:  The server has died.\n");
	goto gendm_done;
      }
    }
    else 
      fprintf(stderr, "Warning: Unable to start calibration.  Continuing.\n");
    
    /* FIXME:  Some servers may require you to restart the server after calibration */
    /* That would be handled here */
  }

  /* From this point, we keep merrily going until the server dies, 
     or auth asks us to stop */

  while(1) {
    int reaped = 0;

    /* Handle the auth (or lack thereof) */

    if (dminfo.flags & FLAG_AUTH) {
      uid_t uid;

      ret = dminfo.authplugin->doauth(&uid);      
      if (ret) goto gendm_done;

      curuser = getpwuid(uid);

      if (!curuser) {
	fprintf(stderr, "Error: user %d does not exist.  Restarting.\n", uid);
	continue;
      }
    }
    else {
      /* If a user was specified, use them, otherwise nothing */

      if (dminfo.user) curuser = dminfo.user;
      else curuser = 0;
    }

    /* If we have no session, then revert back to the default */
    if (!session) session = PDM_DEFAULT_SESSION;

    if (!timestamp || ((time(0) - timestamp) > 30)) {
      timestamp = time(0);
      count = 0;
    }

    /* Start the session */
    VERBOSE("Spawning session '%s'\n", session);
    session_pid = spawn_session(session, curuser ? curuser->pw_uid : getuid());
    
    if (session_pid == -1) {
      fprintf(stderr, "Fatal:  Unable to spawn the session '%s'.\n", session);
      goto gendm_done;
    }

    reaped = wait_for_child();

    /* Welcome back - we've reaped a pid, lets see what we got */
    /* If the server died, then just bail */

    if (reaped == server_pid) {
      fprintf(stderr, "Fatal:  The server has died.\n");
      goto gendm_done;
    }

    if (++count == 5) {
      fprintf(stderr, 
	      "Fatal:  Respawned session '%s' spawned too many times.\n",
	      session);

      goto gendm_done;
    }

    session_pid = 0;
  }
  
 gendm_done:
  if (session_pid > 0) kill_child(session_pid);
  if (server_pid > 0) kill_child(server_pid);

  printf("PDM exiting.  Goodbye.\n");
  return 0;
}
